#

######### test_rwrock.py #########


import threading
import time
from  rwrlock import RWRLock


def writer(L, value, after, rwlock, times):
    """Append value to L after a period of time."""
    try:
        with rwlock.w_locked():
        # Get another lock, to test the fact that obtaining multiple
        # write locks from the same thread context doesn't block (lock
        # reentrancy).
            with rwlock.w_locked():
        # Get a reader lock too; should be the same as getting another
        # writer since writers are inherently readers as well.
                with rwlock.r_locked():
                    times.append(time.time())
                    time.sleep(after)
                    L.append(value)
    finally:
        times.append(time.time())


def reader(L1, L2, after, rwlock, times):
    """Append values from L1 to L2 after a period of time."""
    try:
        with rwlock.r_locked():
        # Get another lock, to test the fact that obtaining multiple
        # write locks from the same thread context doesn't block (lock
        # reentrancy).
            with rwlock.r_locked():
                times.append(time.time())
                time.sleep(after)
                L2.extend(L1)
    finally:
        times.append(time.time())


def readerTurnedWriter(L, value, after, rwlock, times):
    """Append value to L after a period of time."""
    try:
        with rwlock.r_locked():
            with rwlock.w_locked():
                times.append(time.time())
                time.sleep(after)
                L.append(value)
    finally:
        times.append(time.time())



def test_reentrancy():
    lock = RWRLock()
    # Reentrant read locks.
    with lock.r_locked():
        with lock.r_locked():
            pass

    # Reentrant write locks.
    with lock.w_locked():
        with lock.w_locked():
            pass

    # Writers are also readers.
    with lock.w_locked():
        with lock.r_locked():
            pass

    assert(1==1)


def test_reentrancy2locks():
    lock = RWRLock()
    lock2 = RWRLock()

    # Reentrant read locks 2locks.
    with lock.r_locked():
        with lock.r_locked():
            with lock2.r_locked():
                with lock2.r_locked():
                    pass

    # Reentrant write locks.
    with lock.w_locked():
        with lock.w_locked():
            with lock2.w_locked():
                with lock2.w_locked():
                    pass

    # Writers are also readers.
    with lock.w_locked():
        with lock.r_locked():
            with lock2.w_locked():
                with lock2.r_locked():
                    pass

    # Writers are also readers 2llocks.
    with lock.w_locked():
        with lock2.r_locked():
            with lock.w_locked():
                with lock2.r_locked():
                    pass

    with lock.w_locked():
        with lock2.r_locked():
            with lock2.w_locked():
                with lock.r_locked():
                    pass
    assert(1==1)
def test_writeReadRead():
    lock = RWRLock()
    W, R1, R2 = [], [], []
    TW, TR1, TR2 = [], [], []
    thread1 = threading.Thread(
        target=writer,
        args=(W, 'foo', 0.2, lock, TW),
        )
    thread2 = threading.Thread(
        target=reader,
        args=(W, R1, 0.2, lock, TR1),
        )
    thread3 = threading.Thread(
        target=reader,
        args=(W, R2, 0.2, lock, TR2),
        )
    thread1.start()
    time.sleep(0.1)
    thread2.start()
    thread3.start()
    time.sleep(0.8)
    assert 'foo' in R1
    assert 'foo' in R2
    assert TR1[0] <= TR2[1]             # Read 1 started during read 2.
    assert TR2[0] <= TR1[1]             # Read 2 started during read 1.
    assert TR1[0] >= TW[1]              # Read 1 started after write.
    assert TR2[0] >= TW[1]              # Read 2 started after write.


def test_writeReadReadWrite():
    lock = RWRLock()
    W, R1, R2 = [], [], []
    TW1, TR1, TR2, TW2 = [], [], [], []
    thread1 = threading.Thread(
        target=writer,
        args=(W, 'foo', 0.3, lock, TW1),
        )
    thread2 = threading.Thread(
        target=reader,
        args=(W, R1, 0.3, lock, TR1),
        )
    thread3 = threading.Thread(
        target=reader,
        args=(W, R2, 0.3, lock, TR2),
        )
    thread4 = threading.Thread(
        target=writer,
        args=(W, 'bar', 0.3, lock, TW2),
        )
    thread1.start()
    time.sleep(0.1)
    thread2.start()
    time.sleep(0.1)
    thread3.start()
    time.sleep(0.1)
    thread4.start()
    time.sleep(1.7)
    assert 'foo' in R1
    assert 'foo' in R2
    assert 'bar' not in R1
    assert 'bar' not in R2
    assert 'bar' in W
    assert TR1[0] <= TR2[1]              # Read 1 started during read 2.
    assert TR2[0] <= TR1[1]              # Read 2 started during read 1.
    assert TR1[0] >= TW1[1]              # Read 1 started after write 1.
    assert TR2[0] >= TW1[1]              # Read 2 started after write 1.
    assert TW2[0] >= TR1[1]              # Write 2 started after read 1.
    assert TW2[0] >= TR2[1]              # Write 2 started after read 2.


def test_writeReadReadtowrite():
    lock = RWRLock()
    W, R1 = [], []
    TW1, TR1, TW2 = [], [], []
    thread1 = threading.Thread(
        target=writer,
        args=(W, 'foo', 0.3, lock, TW1),
        )
    thread2 = threading.Thread(
        target=reader,
        args=(W, R1, 0.3, lock, TR1),
        )
    thread3 = threading.Thread(
        target=readerTurnedWriter,
        args=(W, 'bar', 0.3, lock, TW2),
        )
    thread1.start()
    time.sleep(0.1)
    thread2.start()
    time.sleep(0.1)
    thread3.start()
    time.sleep(1.7)
    assert 'foo' in R1
    assert 'bar' not in R1
    assert 'bar' in W
    assert TR1[0] >= TW1[1]              # Read 1 started after write 1.
    assert TW2[0] >= TR1[1]              # Write 2 started after read 1.



# -*- coding: utf-8 -*-
""" rwlock.py

    A class to implement read-write locks on top of the standard threading
    library.

    This is implemented with two mutexes (threading.Lock instances) as per this
    wikipedia pseudocode:

    https://en.wikipedia.org/wiki/Readers%E2%80%93writer_lock#Using_two_mutexes

    Code written by Mike Moore.
    
    Based upon code by Tyler Neylon at Unbox Research.
    
    see this gist https://gist.github.com/tylerneylon/a7ff6017b7a1f9a506cf75aa23eacfd6
    
    Added that itappears as a reentrant locks and reference countingto make rw reentrant
"""


# _______________________________________________________________________
# Imports

from contextlib import contextmanager
import threading


# _______________________________________________________________________
# Class

class RWRLock(object):
    """ RWLock class; this is meant to allow an object to be read from by
        multiple threads, but only written to by a single thread at a time. See:
        https://en.wikipedia.org/wiki/Readers%E2%80%93writer_lock

        Usage:

            from rwlock import RWLock

            my_obj_rwlock = RWLock()

            # When reading from my_obj:
            with my_obj_rwlock.r_locked():
                do_read_only_things_with(my_obj)

            # When writing to my_obj:
            with my_obj_rwlock.w_locked():
                mutate(my_obj)
    """

    def __init__(self):
        # as possible to have manyrw locks
        # created per class as lock is expected
        # to be used across many threads
        self._threadlocal = threading.local()

        # has to be a lock and non reentrant as locks can be acquired
        # in one threadand released in another
        self.w_lock = threading.Lock()
        self.num_r_lock = threading.RLock()
        self.num_r = 0
        self.num_w =0
        self.promoteacquire = False

    # ___________________________________________________________________
    # Reading methods.
    def thread_lock_count(self):
        rlockcount = getattr(self._threadlocal, 'rlockcount', None)
        wlockcount = getattr(self._threadlocal, 'wlockcount', None)
        if rlockcount is None:
            rlockcount = 0
        if wlockcount is None:
            wlockcount = 0
        return(rlockcount,wlockcount)

    def set_thread_lock_count(self,rlockcount, wlockcount):
        self._threadlocal.rlockcount =  rlockcount
        self._threadlocal.wlockcount =  wlockcount

    def nr_acquire(self):
        self.num_r_lock.acquire()
        self.num_r += 1
        if self.num_r == 1:
            self.w_lock.acquire()
            self.num_w += 1
        self.num_r_lock.release()

    def r_acquire(self):
        rlockcount,wlockcount = self.thread_lock_count()
        if rlockcount==0 and wlockcount==0:
            self.nr_acquire()
        rlockcount=rlockcount+1
        self.set_thread_lock_count(rlockcount, wlockcount)

    def nr_release(self):
        assert self.num_r > 0
        self.num_r_lock.acquire()
        self.num_r -= 1
        if self.num_r == 0:
            self.num_w -= 1
            self.w_lock.release()
        self.num_r_lock.release()

    def r_release(self):
        rlockcount,wlockcount = self.thread_lock_count()
        assert  rlockcount > 0
        rlockcount = rlockcount - 1
        if rlockcount == 0 and wlockcount==0:
            self.nr_release()
        self.set_thread_lock_count(rlockcount, wlockcount)

    @contextmanager
    def r_locked(self):
        """ This method is designed to be used via the `with` statement. """
        self.r_acquire()
        yield
        self.r_release()

    # ___________________________________________________________________
    # Writing methods.
    def nw_acquire(self):
        # as any one reader ensures in any threadis still running this lock
        # wi notbe obrainable so we wi block
        # if this thread is onlly one running it willl work as arentrant lock
        self.w_lock.acquire()
        self.num_w += 1

    def w_acquire(self):
        rlockcount, wlockcount = self.thread_lock_count()
        if wlockcount == 0:
            # if we havea read lock
            # we remove readlock
            # then getawrite llock
            # nb this means data beforeand afterwrite may change
            # in 1 thread but nit in an inconsistentfashion
            # means non repeated reads for sharedstate
            if rlockcount > 0:
                self.nr_release()
            self.nw_acquire()

        wlockcount = wlockcount + 1
        self.set_thread_lock_count(rlockcount, wlockcount)

    def nw_release(self):
        self.num_w -= 1
        self.w_lock.release()

    def w_release(self):
        rlockcount, wlockcount = self.thread_lock_count()
        assert wlockcount > 0
        wlockcount = wlockcount - 1
        if wlockcount == 0:
            # put back the fact we are a reader
            # if wearea reader i.e. demote write ock back to read
            # if we had no ock before do nothing
            self.nw_release()

            if rlockcount > 0:
                self.nr_acquire()

        self.set_thread_lock_count(rlockcount, wlockcount)

    @contextmanager
    def w_locked(self):
        """ This method is designed to be used via the `with` statement. """
        self.w_acquire()
        yield
        self.w_release()


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
     name='rwrlock',
     version='0.0.2',
     scripts=[] ,
     author="Mike Moore",
     author_email="z_z_zebra@yahoo.com",
     description="A limited re-entrant readwrite lock",
     long_description_content_type='text/markdown',
     long_description=long_description,
     url="https://github.com/Mikemoore63/rwrlock",
     packages=setuptools.find_packages(),
     test_suite='nose.collector',
     use_2to3=True,
     tests_require=['nose'],
     include_package_data=True,
     install_requires=[

      ],
     classifiers=[
         "Programming Language :: Python :: 2",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
     ],

 )
